/**
 * Copyright (c) 2012 - 2017, Nordic Semiconductor ASA
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 * 
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 * 
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 * 
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
/**@cond To Make Doxygen skip documentation generation for this file.
 * @{
 */

#include "sdk_common.h"
#include "ble_mgmt_c.h"
#include "ble_db_discovery.h"
#include "ble_types.h"
#include "ble_srv_common.h"
#include "ble_gattc.h"
//#include "ble_fas.h"
//#include "ble_common.h"

#define NRF_LOG_MODULE_NAME ble_mgmt_c
#include "nrf_log.h"
NRF_LOG_MODULE_REGISTER();


#define TX_BUFFER_MASK         0x07                  /**< TX Buffer mask, must be a mask of continuous zeroes, followed by continuous sequence of ones: 000...111. */
#define TX_BUFFER_SIZE         (TX_BUFFER_MASK + 1)  /**< Size of send buffer, which is 1 higher than the mask. */

#define WRITE_MESSAGE_LENGTH   BLE_CCCD_VALUE_LEN    /**< Length of the write message for CCCD. */
#define WRITE_MESSAGE_LENGTH   BLE_CCCD_VALUE_LEN    /**< Length of the write message for CCCD. */


typedef enum
{
    READ_REQ,  /**< Type identifying that this tx_message is a read request. */
    WRITE_REQ  /**< Type identifying that this tx_message is a write request. */
} tx_request_t;

/**@brief Structure for writing a message to the peer, i.e. CCCD.
 */
typedef struct
{
    uint8_t                  gattc_value[WRITE_MESSAGE_LENGTH];  /**< The message to write. */
    ble_gattc_write_params_t gattc_params;                       /**< GATTC parameters for this message. */
} write_params_t;

/**@brief Structure for holding data to be transmitted to the connected central.
 */
typedef struct
{
    uint16_t     conn_handle;  /**< Connection handle to be used when transmitting this message. */
    tx_request_t type;         /**< Type of this message, i.e. read or write message. */
    union
    {
        uint16_t       read_handle;  /**< Read request message. */
        write_params_t write_req;    /**< Write request message. */
    } req;
} tx_message_t;


static tx_message_t  m_tx_buffer[TX_BUFFER_SIZE];  /**< Transmit buffer for messages to be transmitted to the central. */
static uint32_t      m_tx_insert_index = 0;        /**< Current index in the transmit buffer where the next message should be inserted. */
static uint32_t      m_tx_index = 0;               /**< Current index in the transmit buffer from where the next message to be transmitted resides. */


/**@brief Function for passing any pending request from the buffer to the stack.
 */
static void tx_buffer_process(void)
{
  if (m_tx_index != m_tx_insert_index)
    {
      uint32_t err_code;

      if (m_tx_buffer[m_tx_index].type == READ_REQ)
        {
          //NRF_LOG_INFO("sd_ble_gattc_read[%d][%d]", m_tx_buffer[m_tx_index].conn_handle, m_tx_buffer[m_tx_index].req.read_handle);
          err_code = sd_ble_gattc_read(m_tx_buffer[m_tx_index].conn_handle,
                                       m_tx_buffer[m_tx_index].req.read_handle,
                                       0);
          if (err_code != NRF_SUCCESS)
            {
              NRF_LOG_INFO("sd_ble_gattc_read error[%d]", err_code);
            }
        }
      else
        {
          //NRF_LOG_INFO("sd_ble_gattc_write[%d][%d]", m_tx_buffer[m_tx_index].conn_handle, m_tx_buffer[m_tx_index].req.write_req.gattc_params.handle);
          err_code = sd_ble_gattc_write(m_tx_buffer[m_tx_index].conn_handle,
                                        &m_tx_buffer[m_tx_index].req.write_req.gattc_params);
          if (err_code != NRF_SUCCESS)
            {
              NRF_LOG_INFO("sd_ble_gattc_write error[%d]", err_code);
            }
        }
      if (err_code == NRF_SUCCESS)
        {
          m_tx_index++;
          m_tx_index &= TX_BUFFER_MASK;
        }
    }
}


/**@brief     Function for handling write response events.
 *
 * @param[in] p_ble_mgmt_c Pointer to the ble management Client structure.
 * @param[in] p_ble_evt   Pointer to the BLE event received.
 */
static void on_write_rsp(ble_mgmt_c_t * p_ble_mgmt_c, const ble_evt_t * p_ble_evt)
{
  const ble_gattc_evt_write_rsp_t * p_response;
    // Check if the event if on the link for this instance
    if (p_ble_mgmt_c->conn_handle != p_ble_evt->evt.gattc_evt.conn_handle)
    {
        return;
    }
    p_response = &p_ble_evt->evt.gattc_evt.params.write_rsp;

    NRF_LOG_INFO("on_write_rsp = evt[%d], hdl[%d]", p_ble_evt->header.evt_id, p_ble_evt->evt.gattc_evt.params.write_rsp.handle);
    if((p_response ->handle == p_ble_mgmt_c->peer_mgmt_db.cvw_handle))
      {
        ble_mgmt_c_evt_t ble_fcs_c_evt;

        ble_fcs_c_evt.conn_handle = p_ble_evt->evt.gattc_evt.conn_handle;
        ble_fcs_c_evt.evt_type = BLE_MGMT_C_EVT_WRITE_COMPLETE;
        p_ble_mgmt_c->evt_handler(p_ble_mgmt_c, &ble_fcs_c_evt);
      }
    // Check if there is any message to be sent across to the peer and send it.
    tx_buffer_process();
}


/**@brief     Function for handling Handle Value Notification received from the SoftDevice.
 *
 * @details   This function will uses the Handle Value Notification received from the SoftDevice
 *            and checks if it is a notification of the ble management measurement from the peer. If
 *            it is, this function will decode the ble management measurement and send it to the
 *            application.
 *
 * @param[in] p_ble_mgmt_c Pointer to the ble management Client structure.
 * @param[in] p_ble_evt   Pointer to the BLE event received.
 */
static void on_hvx(ble_mgmt_c_t * p_ble_mgmt_c, const ble_evt_t * p_ble_evt)
{
}

/**@brief     Function for handling read response events.
 *
 * @details   This function will validate the read response and raise the appropriate
 *            event to the application.
 *
 * @param[in] p_ble_mgmt_c Pointer to the ble management Client structure.
 * @param[in] p_ble_evt   Pointer to the BLE event received.
 */
static void on_read_rsp(ble_mgmt_c_t * p_ble_mgmt_c, const ble_evt_t * p_ble_evt)
{
    const ble_gattc_evt_read_rsp_t * p_response;

    NRF_LOG_INFO("on_read_rsp = evt[%d], hdl[%d], conn[%d]", p_ble_evt->header.evt_id, p_ble_evt->evt.gattc_evt.params.read_rsp.handle, p_ble_evt->evt.gattc_evt.conn_handle);
    NRF_LOG_INFO("on_read_rsp = hdl[%d], conn[%d]",p_ble_mgmt_c->peer_mgmt_db.rnc_handle, p_ble_mgmt_c->conn_handle);
    // Check if the event if on the link for this instance
    if (p_ble_mgmt_c->conn_handle != p_ble_evt->evt.gattc_evt.conn_handle)
    {
        return;
    }

    p_response = &p_ble_evt->evt.gattc_evt.params.read_rsp;

    if (p_response->handle == p_ble_mgmt_c->peer_mgmt_db.rnc_handle)
    {
        ble_mgmt_c_evt_t ble_mgmt_c_evt;

        ble_mgmt_c_evt.conn_handle = p_ble_evt->evt.gattc_evt.conn_handle;
        ble_mgmt_c_evt.evt_type = BLE_MGMT_C_EVT_RANDOM_NUMBER_READ;

        ble_mgmt_c_evt.params.rnc = p_response->data[0];
        ble_mgmt_c_evt.params.rnc = (ble_mgmt_c_evt.params.rnc << 8) + p_response->data[1];
        NRF_LOG_INFO("rnc[0x%x]",ble_mgmt_c_evt.params.rnc);

        p_ble_mgmt_c->evt_handler(p_ble_mgmt_c, &ble_mgmt_c_evt);
    }
   // Check if there is any buffered transmissions and send them.
    tx_buffer_process();
}

/**@brief     Function for handling Disconnected event received from the SoftDevice.
 *
 * @details   This function check if the disconnect event is happening on the link
 *            associated with the current instance of the module, if so it will set its
 *            conn_handle to invalid.
 *
 * @param[in] p_ble_mgmt_c Pointer to the ble management Client structure.
 * @param[in] p_ble_evt   Pointer to the BLE event received.
 */
static void on_disconnected(ble_mgmt_c_t * p_ble_mgmt_c, const ble_evt_t * p_ble_evt)
{
    if (p_ble_mgmt_c->conn_handle == p_ble_evt->evt.gap_evt.conn_handle)
    {
        p_ble_mgmt_c->conn_handle                 = BLE_CONN_HANDLE_INVALID;
        p_ble_mgmt_c->peer_mgmt_db.mgmt_cccd_handle = BLE_GATT_HANDLE_INVALID;
        p_ble_mgmt_c->peer_mgmt_db.rnc_handle      = BLE_GATT_HANDLE_INVALID;
        p_ble_mgmt_c->peer_mgmt_db.cvw_handle      = BLE_GATT_HANDLE_INVALID;
    }
}


void ble_mgmt_on_db_disc_evt(ble_mgmt_c_t * p_ble_mgmt_c, const ble_db_discovery_evt_t * p_evt)
{
    // Check if the ble management Service was discovered.
    if (p_evt->evt_type == BLE_DB_DISCOVERY_COMPLETE &&
        p_evt->params.discovered_db.srv_uuid.uuid == BLE_FAS_BASE_UUID &&
        p_evt->params.discovered_db.srv_uuid.type == BLE_UUID_TYPE_VENDOR_BEGIN)
    {
        // Find the CCCD Handle of the ble management Measurement characteristic.
        uint32_t i;

        ble_mgmt_c_evt_t evt;

        evt.evt_type    = BLE_MGMT_C_EVT_DISCOVERY_COMPLETE;
        evt.conn_handle = p_evt->conn_handle;

        for (i = 0; i < p_evt->params.discovered_db.char_count; i++)
        {
            if (p_evt->params.discovered_db.charateristics[i].characteristic.uuid.uuid ==
                BLE_FAS_UUID_RAND_NUM_CLR) //! Random number read
            {
                // Found ble management characteristic. Store CCCD handle and break.
                evt.params.peer_db.mgmt_cccd_handle =
                    p_evt->params.discovered_db.charateristics[i].cccd_handle;
                evt.params.peer_db.rnc_handle =
                    p_evt->params.discovered_db.charateristics[i].characteristic.handle_value;
                //break;
            }
            if (p_evt->params.discovered_db.charateristics[i].characteristic.uuid.uuid ==
                BLE_FAS_UUID_CIPHOR_WRITE)
            {
                evt.params.peer_db.cvw_handle =
                    p_evt->params.discovered_db.charateristics[i].characteristic.handle_value;
                //break;
            }
        }

        NRF_LOG_INFO("FAS Service discovered at peer.");
        //If the instance has been assigned prior to db_discovery, assign the db_handles
        if (p_ble_mgmt_c->conn_handle != BLE_CONN_HANDLE_INVALID)
        {
            if ((p_ble_mgmt_c->peer_mgmt_db.mgmt_cccd_handle == BLE_GATT_HANDLE_INVALID)&&
                (p_ble_mgmt_c->peer_mgmt_db.rnc_handle == BLE_GATT_HANDLE_INVALID))
            {
                p_ble_mgmt_c->peer_mgmt_db = evt.params.peer_db;
            }
        }


        p_ble_mgmt_c->evt_handler(p_ble_mgmt_c, &evt);
    }
}


uint32_t ble_mgmt_c_init(ble_mgmt_c_t * p_ble_mgmt_c, ble_mgmt_c_init_t * p_ble_mgmt_c_init)
{
    VERIFY_PARAM_NOT_NULL(p_ble_mgmt_c);
    VERIFY_PARAM_NOT_NULL(p_ble_mgmt_c_init);

    ble_uuid_t mgmt_uuid;

    mgmt_uuid.type = BLE_UUID_TYPE_VENDOR_BEGIN ; //BLE_UUID_TYPE_BLE; //


    p_ble_mgmt_c->evt_handler                 = p_ble_mgmt_c_init->evt_handler;
    p_ble_mgmt_c->conn_handle                 = BLE_CONN_HANDLE_INVALID;
    p_ble_mgmt_c->peer_mgmt_db.mgmt_cccd_handle = BLE_GATT_HANDLE_INVALID;
    //p_ble_mgmt_c->peer_mgmt_db.rnc_handle      = BLE_GATT_HANDLE_INVALID;
    //p_ble_mgmt_c->peer_mgmt_db.cvw_handle      = BLE_GATT_HANDLE_INVALID;

    mgmt_uuid.uuid = BLE_FAS_BASE_UUID; // discover fas service
    return ble_db_discovery_evt_register(&mgmt_uuid);
}

void ble_mgmt_c_on_ble_evt(ble_evt_t const * p_ble_evt, void * p_context)
{
    ble_mgmt_c_t * p_ble_mgmt_c = (ble_mgmt_c_t *)p_context;

    //NRF_LOG_INFO("p_ble_evt->header.evt_id -> %d",p_ble_evt->header.evt_id);
   if ((p_ble_mgmt_c == NULL) || (p_ble_evt == NULL))
    {
        return;
    }


    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GATTC_EVT_HVX:
            on_hvx(p_ble_mgmt_c, p_ble_evt);
            break;

        case BLE_GATTC_EVT_READ_RSP:
            on_read_rsp(p_ble_mgmt_c, p_ble_evt);
            break;

        case BLE_GATTC_EVT_WRITE_RSP:
            on_write_rsp(p_ble_mgmt_c, p_ble_evt);
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            on_disconnected(p_ble_mgmt_c, p_ble_evt);
            break;

        default:
            break;
    }
}




uint32_t ble_mgmt_c_handles_assign(ble_mgmt_c_t * p_ble_mgmt_c,
                                  uint16_t conn_handle,
                                  const mgmt_db_t * p_peer_mgmt_handles)
{
    VERIFY_PARAM_NOT_NULL(p_ble_mgmt_c);

    p_ble_mgmt_c->conn_handle = conn_handle;
    if (p_peer_mgmt_handles != NULL)
    {
        p_ble_mgmt_c->peer_mgmt_db = *p_peer_mgmt_handles;
    }
    return NRF_SUCCESS;
}

/**@brief Function for creating a message for reading random number.
 */
uint32_t ble_mgmt_c_read_random_number(ble_mgmt_c_t * p_ble_mgmt_c)
{
    tx_message_t * p_msg;

    p_msg              = &m_tx_buffer[m_tx_insert_index++];
    m_tx_insert_index &= TX_BUFFER_MASK;

    p_msg->req.read_handle                 = p_ble_mgmt_c -> peer_mgmt_db.rnc_handle;
    p_msg->conn_handle                     = p_ble_mgmt_c -> conn_handle;
    p_msg->type                            = READ_REQ;

    NRF_LOG_INFO("Reading Random Number Handle = %d", p_msg->req.read_handle);
    tx_buffer_process();
    return NRF_SUCCESS;
}

/**@brief Function for creating a message for writing cipher value.
 */
uint32_t ble_mgmt_c_write_cipher_value(ble_mgmt_c_t * p_mgmt_c, ble_mgmt_c_evt_t * p_mgmt_c_evt)
{

    tx_message_t * p_msg;
    p_msg              = &m_tx_buffer[m_tx_insert_index++];
    m_tx_insert_index &= TX_BUFFER_MASK;

    p_msg->req.write_req.gattc_params.handle   = p_mgmt_c->peer_mgmt_db.cvw_handle;
    p_msg->req.write_req.gattc_params.len      = sizeof(p_mgmt_c_evt->params.cvw);
    p_msg->req.write_req.gattc_params.p_value  = p_msg->req.write_req.gattc_value;
    p_msg->req.write_req.gattc_params.offset   = 0;
    p_msg->req.write_req.gattc_params.write_op = BLE_GATT_OP_WRITE_REQ;
    p_msg->req.write_req.gattc_value[0]        = LSB_16(p_mgmt_c_evt->params.cvw);
    p_msg->req.write_req.gattc_value[1]        = MSB_16(p_mgmt_c_evt->params.cvw);
    p_msg->conn_handle                         = p_mgmt_c_evt->conn_handle;
    p_msg->type                                = WRITE_REQ;

    NRF_LOG_INFO("write_cipher_value Handle = %d", p_msg->req.write_req.gattc_params.handle);
    tx_buffer_process();
    return NRF_SUCCESS;
}

/**@brief Function for creating a message for writing by ble handle.
 */
uint32_t ble_mgmt_char_write_by_handle(uint16_t conn_handle, uint16_t handle, uint8_t * data, uint8_t len)
{
  uint32_t err_code;
  tx_message_t msg;
  uint8_t index;

  msg.req.write_req.gattc_params.handle   = handle;
  msg.req.write_req.gattc_params.len      = len;
  msg.req.write_req.gattc_params.p_value  = msg.req.write_req.gattc_value;
  msg.req.write_req.gattc_params.offset   = 0;
  msg.req.write_req.gattc_params.write_op = BLE_GATT_OP_WRITE_REQ;
  for ( index = 0 ; index < len ; index++ )
    {
      msg.req.write_req.gattc_value[index]        = data[index];
    }
  msg.conn_handle                         = conn_handle;
  msg.type                                = WRITE_REQ;

  err_code = sd_ble_gattc_write(msg.conn_handle,
                                &msg.req.write_req.gattc_params);
  return err_code;
}

/**@brief Function for creating a message for reading by ble handle.
 */
uint32_t ble_mgmt_char_read_by_handle(uint16_t conn_handle, uint16_t handle)
{
    tx_message_t * p_msg;

    p_msg              = &m_tx_buffer[m_tx_insert_index++];
    m_tx_insert_index &= TX_BUFFER_MASK;

    p_msg->req.read_handle                 = handle;
    p_msg->conn_handle                     = conn_handle;
    p_msg->type                            = READ_REQ;

    tx_buffer_process();
    return NRF_SUCCESS;
}

/**@brief Function for creating a message for writing by ble handle.
 */
uint32_t ble_common_write_by_handle(uint16_t conn_handle, uint16_t handle, uint8_t * data, uint8_t len)
{
  uint32_t err_code;
  tx_message_t msg;

  msg.req.write_req.gattc_params.handle   = handle;
  msg.req.write_req.gattc_params.len      = len;
  msg.req.write_req.gattc_params.p_value  = data;
  msg.req.write_req.gattc_params.offset   = 0;
  msg.req.write_req.gattc_params.write_op = BLE_GATT_OP_WRITE_REQ;

  DataUtil_changeEndianBuffer(data, len);
  memcpy(msg.req.write_req.gattc_value, data, len);
  msg.conn_handle                         = conn_handle;
  msg.type                                = WRITE_REQ;

  err_code = sd_ble_gattc_write(msg.conn_handle,
                                &msg.req.write_req.gattc_params);
  NRF_LOG_INFO("Write by Handle %d, error[%d]",handle, err_code);
  return err_code;
}

/**@brief Function for creating a message for reading random number.
 */
uint32_t ble_common_read_by_handle(uint16_t conn_handle, uint16_t handle)
{
  uint32_t err_code;

  err_code = sd_ble_gattc_read(conn_handle,
                               handle,
                               0);
  NRF_LOG_INFO("Reading by Handle %d, error[%d]",handle, err_code);
  return err_code;
}

/** @}
 *  @endcond
 */

